/*
    This file is a part of the SteamWar software.

    Copyright (C) 2022  SteamWar.de-Serverteam

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.steamwar.fightsystem.utils;

import com.comphenix.tinyprotocol.Reflection;
import com.comphenix.tinyprotocol.TinyProtocol;
import de.steamwar.core.Core;
import de.steamwar.entity.REntity;
import de.steamwar.fightsystem.Config;
import de.steamwar.fightsystem.fight.Fight;
import de.steamwar.fightsystem.fight.FightPlayer;
import de.steamwar.fightsystem.fight.FightTeam;
import de.steamwar.fightsystem.listener.Recording;
import de.steamwar.fightsystem.states.FightState;
import de.steamwar.fightsystem.states.StateDependent;
import de.steamwar.fightsystem.states.StateDependentListener;
import de.steamwar.fightsystem.states.StateDependentTask;
import de.steamwar.techhider.TechHider;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockPhysicsEvent;
import org.bukkit.event.entity.EntityDeathEvent;
import org.bukkit.event.entity.EntitySpawnEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.event.player.PlayerQuitEvent;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.BiFunction;
import java.util.function.Function;

public class HullHider implements Listener {

    private final Map<FightTeam, Hull> hullMap = new HashMap<>();
    private final Hull[] hulls;
    private final Map<Class<?>, BiFunction<Player, Object, Object>> packetHiders = new HashMap<>();

    public HullHider() {
        if(!TechHiderWrapper.ENABLED) {
            hulls = new Hull[0];
            return;
        }

        Fight.teams().forEach(team -> hullMap.put(team, new Hull(team)));
        hulls = hullMap.values().toArray(new Hull[0]);

        packetHiders.put(packetPlayOutWorldEvent, this::worldEventHider);
        packetHiders.put(packetPlayOutExplosion, this::explosionHider);
        posHiderGenerator("{nms.network.protocol.game}.PacketPlayOutWorldParticles", Core.getVersion() >= 18 ? double.class : float.class, 1.0);
        posHiderGenerator("{nms.network.protocol.game}.PacketPlayOutNamedSoundEffect", int.class, 8.0);
        if(Core.getVersion() >= 9 && Core.getVersion() < 18)
            posHiderGenerator("{nms.network.protocol.game}.PacketPlayOutCustomSoundEffect", int.class, 8.0);

        new StateDependentListener(TechHiderWrapper.ENABLED, FightState.Schem, this);
        new StateDependent(TechHiderWrapper.ENABLED, FightState.Schem) {
            @Override
            public void enable() {
                packetHiders.forEach(TinyProtocol.instance::addFilter);
                Bukkit.getOnlinePlayers().forEach(HullHider.this::updatePlayer);
            }

            @Override
            public void disable() {
                Bukkit.getOnlinePlayers().forEach(player -> removePlayer(player, true));
                packetHiders.forEach(TinyProtocol.instance::removeFilter);
            }
        }.register();
        new StateDependentTask(TechHiderWrapper.ENABLED, FightState.Schem, this::onTick, 0, 1);
    }

    public void initialize(FightTeam team) {
        if(!TechHiderWrapper.ENABLED)
            return;

        hullMap.get(team).initialize();
    }


    @EventHandler(priority = EventPriority.HIGH)
    public void onJoin(PlayerJoinEvent e) {
        updatePlayer(e.getPlayer());
    }

    @EventHandler
    public void onLeave(PlayerQuitEvent e) {
        removePlayer(e.getPlayer(), false);
    }

    public void updatePlayer(Player player) {
        if(!TechHiderWrapper.ENABLED)
            return;

        FightTeam team = Fight.getPlayerTeam(player);
        FightPlayer fp = Fight.getFightPlayer(player);
        for(Map.Entry<FightTeam, Hull> hull : hullMap.entrySet()) {
            if(Config.isReferee(player) || hull.getKey() == team || (fp != null && fp.canEntern())) {
                hull.getValue().removePlayer(player, true);
            } else {
                hull.getValue().addPlayer(player);
            }
        }
    }

    private void removePlayer(Player player, boolean activeRemoval) {
        for (Hull hull : hulls)
            hull.removePlayer(player, activeRemoval);
    }


    @EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
    public void onBlockPhysic(BlockPhysicsEvent e) {
        if(FlatteningWrapper.impl.doRecord(e))
            blockUpdate(e.getBlock(), e.getChangedType());
    }

    public void blockUpdate(Block block, Material changedType) {
        for (Hull hull : hulls)
            hull.updateBlockVisibility(block, changedType);
    }

    public boolean isBlockHidden(Player player, int x, int y, int z) {
        if(!TechHiderWrapper.ENABLED)
            return false;

        for (Hull hull : hulls)
            if(hull.isBlockHidden(player, x, y, z))
                return true;

        return false;
    }

    public boolean blockPrecise(Player player, int chunkX, int chunkY, int chunkZ) {
        if(!TechHiderWrapper.ENABLED)
            return false;

        for (Hull hull : hulls)
            if(hull.blockPrecise(player, chunkX, chunkY, chunkZ))
                return true;

        return false;
    }

    @EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
    public void onSpawn(EntitySpawnEvent e) {
        for (Hull hull : hulls)
            hull.checkEntity(e.getEntity());
    }

    private void onTick() {
        Recording.iterateOverEntities(Objects::nonNull, entity -> {
            for (Hull hull : hulls)
                hull.checkEntity(entity);
        });

        for (Hull hull : hulls)
            hull.sendUncoveredBlocks();
    }

    @EventHandler(priority = EventPriority.MONITOR, ignoreCancelled = true)
    public void onDeath(EntityDeathEvent e) {
        for(Hull hull : hulls)
            hull.removeEntity(e.getEntity());
    }

    public void updateREntity(REntity e) {
        for(Hull hull : hulls)
            hull.checkREntity(e);
    }

    public void despawnREntity(REntity e) {
        for(Hull hull : hulls)
            hull.removeREntity(e);
    }


    private static final Class<?> packetPlayOutWorldEvent = Reflection.getClass("{nms.network.protocol.game}.PacketPlayOutWorldEvent");
    private static final Reflection.FieldAccessor<?> worldEventPosition = Reflection.getField(packetPlayOutWorldEvent, TechHider.blockPosition, 0);
    public static final Reflection.FieldAccessor<Integer> blockPositionY = Reflection.getField("{nms.core}.BaseBlockPosition", int.class, 1);
    private Object worldEventHider(Player player, Object packet) {
        Object baseBlock = worldEventPosition.get(packet);
        return packetHider(player, packet, new Location(Config.world, TechHider.blockPositionX.get(baseBlock), blockPositionY.get(baseBlock), TechHider.blockPositionZ.get(baseBlock)));
    }

    private static final Class<?> packetPlayOutExplosion = Reflection.getClass("{nms.network.protocol.game}.PacketPlayOutExplosion");
    private static final Reflection.FieldAccessor<List> explosionBlocks = Reflection.getField(packetPlayOutExplosion, List.class, 0);
    private static final Function<Object, Location> explosionLocation = posPacketToLocation(packetPlayOutExplosion, double.class, 1.0);
    private Object explosionHider(Player player, Object packet) {
        if(explosionBlocks.get(packet).isEmpty())
            return packetHider(player, packet, explosionLocation.apply(packet));

        return packet;
    }

    private void posHiderGenerator(String typeName, Class<? extends Number> posType, double factor) {
        Class<?> type = Reflection.getClass(typeName);
        Function<Object, Location> location = posPacketToLocation(type, posType, factor);
        packetHiders.put(type, (player, packet) -> packetHider(player, packet, location.apply(packet)));
    }

    private static Function<Object, Location> posPacketToLocation(Class<?> type, Class<? extends Number> posType, double factor) {
        Reflection.FieldAccessor<? extends Number> x = Reflection.getField(type, posType, 0);
        Reflection.FieldAccessor<? extends Number> y = Reflection.getField(type, posType, 1);
        Reflection.FieldAccessor<? extends Number> z = Reflection.getField(type, posType, 2);

        return packet -> new Location(Config.world, x.get(packet).doubleValue()/factor, y.get(packet).doubleValue()/factor, z.get(packet).doubleValue()/factor);
    }

    private Object packetHider(Player player, Object packet, Location location) {
        for(Hull hull : hulls) {
            if(hull.isLocationHidden(player, location))
                return null;
        }

        return packet;
    }
}
